{ TCrossEdit version 3.0
  Udo Juerss, May 1999
  57078 Siegen, Germany
  e-mail: ujhs@aol.com

  CrossEdit version 3.0 is a versatile single-line TEdit
    EditType: String, Float, Integer, Hexadecimal, Binary
    StringCharSet: scFull,scAlpha,scDigit,scTime,scDate
    Precision: float conversion
    Decimals: float conversion
    NextDialogOnEnter: enables focusing next dialog with enter-key
    DialogOnCursorKeys: enables focusing prev/next dialog with up/down keys
    AutoFocus: sets focus if mouse moves over
    FocusAlignment: left, center, right
    FocusColor: color on focus
    FocusWidthInc: expands or shrinks control if it gets the focus
    NoFocusAlignment: left, center, right
    NoFocusColor: color on no focus
    ErrorColor: color on limit- and conversionerror
    LimitCheck: switch for limitcheck
    Min: min value of input
    Max: max value of input
    OnLimitMin: event on limit lower Min
    OnLimitMax: event on limit higher Max
    OnInvalidValue: event on invalid value

  Outfits:  With "AutoFocus" you can enable the control getting focused if the
            mouse moves over.

            Different alignments can be done with "FocusAlignment" and
            "NoFocusAlignment".

            With "FocusWidthInc" <> 0 the control can expand or shrink
            (negative value) if it gets the focus.

            With "FocusColor", "NoFocusColor" you can select different Background
            Colors. "ErrorColor" gets active if the focus is lost with no valid
            Value.

  Function: With "EditType" you can select between String, Integer, Float,
            Binary and HexaDecimal conversion. Even if Binary or Hex is select,
            you can read or write the control with an integer value by using
            GetInteger/SetInteger. "MaxLength" is used to control the length of
            Binary/Hex inputs or outputs. The strings of binary or hex inputs
            can be read with "Text".

            If Float is selected then "Precision" and "Decimals" control the
            conversion of the output string. The actual DecimalSeparator is used.

            "NextDialogOnEnter" enables wether the next control is focused by
            pressing Enter or not.

            "DialogOnCursorKeys" enables wether the next or previous control
            is focused by pressing Up or Down key.
            (Only with Win98 and WinNT5, look at CreateParams)

         -| With "Min" and "Max" you can define limits. If the input is off limits,
         |  an error condition is set and the background gets "ErrorColor".
         |
         -| You can assign "OnLimitMin", "OnLimitMax" to act on these errors.
         |  "OnInvalidValue" signals a conversion error.
         |
         -> This works only if "LimitCheck" is set to True.

            With VK_PRIOR (PgUp) and VK_NEXT (PgDn) you can increment/decrement
            the inputvalue by "NextPriorStep" (only in range Min..Max).

            If you set EditType to etString, then you can set StringCharSet to
            (scFull,scAlpha,scDigit,scTime,scDate). This is useful if you want
            a string but allow only certain characters.

            For most flexible operation there is no other internal errorhandling.

            Public IO-methods:

            procedure SetInteger(Value: Integer);
            procedure SetFloat(Value: Double);
            function GetInteger: Integer;
            function GetFloat: Double;
}

unit
  CrossEdit;

interface
{------------------------------------------------------------------------------}

uses
  Windows,Messages,Classes,Graphics,Controls,StdCtrls;
{------------------------------------------------------------------------------}

const
  NumChars: set of Char = ['0'..'9'];
  MaxBinary = 32;
{------------------------------------------------------------------------------}

type
  TEditType = (etBinary,etFloat,etHex,etInteger,etString);
  TStringCharSet = (scFull,scAlpha,scDigit,scTime,scDate);

  TCrossEdit = class(TEdit)
  private
    FEditType: TEditType;
    FAutoFocus: Boolean;
    FNextDialogOnEnter: Boolean;
    FDialogOnCursorKeys: Boolean;
    FNextPriorStep: Integer;
    FPrecision: Byte;
    FDecimals: Byte;
    FFocusWidthInc: Integer;
    FMin: Double;
    FMax: Double;
    FFocusColor: TColor;
    FNoFocusColor: TColor;
    FErrorColor: TColor;
    FFocusAlignment: TAlignment;
    FNoFocusAlignment: TAlignment;
    FLimitCheck: Boolean;
    FStringCharSet: TStringCharSet;
    FOnInvalidValue: TNotifyEvent;
    FOnLimitMin: TNotifyEvent;
    FOnLimitMax: TNotifyEvent;

    Alignment: TAlignment;
    CharSet: set of Char;
    Error: Boolean;
    KeyPressed: Boolean;
    IsLoaded: Boolean;
  protected
    procedure KeyPress(var Key: Char); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure WndProc(var Message: TMessage); override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure Loaded; override;
    procedure Change; override;
    procedure FormatValue;
    function NoValidValue: Boolean;
    procedure SetAlignment(Value: TAlignment);
    procedure SetNoFocusAlignment(Value: TAlignment);
    procedure SetEditType(Value: TEditType);
    procedure SetColor(Index: Integer; Value: TColor);
    procedure SetStringCharSet(Value: TStringCharSet);
    function IntToBin(Value,Digits: Integer): ShortString;
    function BinToInt(const Value: ShortString): Integer;
  public
    constructor Create(AOwner: TComponent); override;
    function CheckError: Boolean;
    procedure SetInteger(Value: Integer);
    procedure SetFloat(Value: Double);
    function GetInteger: Integer;
    function GetFloat: Double;
  published
    property FocusAlignment: TAlignment read FFocusAlignment write FFocusAlignment;
    property NoFocusAlignment: TAlignment read FNoFocusAlignment write SetNoFocusAlignment;
    property Precision: Byte index 0 read FPrecision write FPrecision;
    property Decimals: Byte index 1 read FDecimals write FDecimals;
    property FocusWidthInc: Integer read FFocusWidthInc write FFocusWidthInc;
    property EditType: TEditType read FEditType write SetEditType;
    property NextDialogOnEnter: Boolean read FNextDialogOnEnter write FNextDialogOnEnter;
    property DialogOnCursorKeys: Boolean read FDialogOnCursorKeys write FDialogOnCursorKeys;
    property NextPriorStep: Integer read FNextPriorStep write FNextPriorStep;
    property AutoFocus: Boolean read FAutoFocus write FAutoFocus;
    property LimitCheck: Boolean read FLimitCheck write FLimitCheck;
    property Min: Double read FMin write FMin;
    property Max: Double read FMax write FMax;
    property FocusColor: TColor index 0 read FFocusColor write SetColor;
    property NoFocusColor: TColor index 1 read FNoFocusColor write SetColor;
    property ErrorColor: TColor index 2 read FErrorColor write SetColor;
    property StringCharSet: TStringCharSet read FStringCharSet write SetStringCharSet;

    property OnInvalidValue: TNotifyEvent read FOnInvalidValue write FOnInvalidValue;
    property OnLimitMin: TNotifyEvent read FOnLimitMin write FOnLimitMin;
    property OnLimitMax: TNotifyEvent read FOnLimitMax write FOnLimitMax;
  end;
{------------------------------------------------------------------------------}

procedure Register;

implementation

uses
  SysUtils,Forms;
{------------------------------------------------------------------------------}

procedure Register;
begin
  RegisterComponents('Udo',[TCrossEdit]);
end;
{------------------------------------------------------------------------------}

constructor TCrossEdit.Create;
begin
  inherited;
  FPrecision:=15;
  FDecimals:=2;
  FNextDialogOnEnter:=True;
  FDialogOnCursorKeys:=True;
  FNextPriorStep:=1;
  FAutoFocus:=True;
  FLimitCheck:=False;
  FFocusColor:=clWhite;
  FNoFocusColor:=clSilver;
  FErrorColor:=clRed;
  FFocusWidthInc:=0;
  FStringCharSet:=scFull;
  FFocusAlignment:=taLeftJustify;
  FNoFocusAlignment:=taRightJustify;
  FMin:=0;
  FMax:=MaxInt;
  FEditType:=etInteger;
  FOnInvalidValue:=nil;
  FOnLimitMin:=nil;
  FOnLimitMax:=nil;

  AutoSelect:=True;
  Height:=13;
  Width:=36;
  BorderStyle:=bsNone;
  Color:=clSilver;
  ParentFont:=False;
  Error:=False;
  Alignment:=taLeftJustify;
  IsLoaded:=False;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  case Alignment of
{ use this with Win95 (DialogOnCursorKeys does not work then)

    taLeftJustify: Params.Style:=Params.Style or ES_LEFT or ES_MULTILINE;
    taRightJustify: Params.Style:=Params.Style or ES_RIGHT or ES_MULTILINE;
    taCenter: Params.Style:=Params.Style or ES_CENTER or ES_MULTILINE;
}
    taLeftJustify: Params.Style:=Params.Style or ES_LEFT;
    taRightJustify: Params.Style:=Params.Style or ES_RIGHT;
    taCenter: Params.Style:=Params.Style or ES_CENTER;
  end;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.DoEnter;
begin
  inherited;
  Error:=False;
  if FFocusWidthInc <> 0 then Width:=Width + FFocusWidthInc;
  Perform(WM_PAINT,0,0);
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.DoExit;
begin
  inherited;
  if FFocusWidthInc <> 0 then Width:=Width - FFocusWidthInc;
  FormatValue;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.Loaded;
begin
  inherited;
  FormatValue;
  SetAlignment(FNoFocusAlignment);
  IsLoaded:=True;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.Change;
begin
  inherited;
  if IsLoaded and not KeyPressed then FormatValue;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.KeyPress(var Key: Char);
begin
  if not (FEditType = etString) then Key:=UpCase(Key);
  if (Key = #13) and FNextDialogOnEnter then
    (Owner as TControl).Perform(WM_NEXTDLGCTL,0,0);
  if (FEditType = etFloat) or (FEditType = etInteger) then
  begin
    if Key = '-' then
    begin
      if Pos('-',Text) > 0 then if SelLength > 0 then Text:='' else Key:=#0;
      if (SelLength = 0) and (SelStart > 0) then Key:=#0;
    end;
    if (Key = DecimalSeparator) and (Pos(DecimalSeparator,Text) > 0) then
      if SelLength > 0 then Text:='' else Key:=#0;
  end;
  if not ((Key = #8) or (Key in CharSet)) then Key:=#0;
  inherited;
  KeyPressed:=True;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.KeyDown(var Key: Word; Shift: TShiftState);
var
  CtlDir: Word;
  I: Integer;
begin
  if FDialogOnCursorKeys and ((Key = VK_UP) or (Key = VK_DOWN)) then
  begin
    if Key = VK_UP then CtlDir:=1 else CtlDir:=0;
    (Owner as TControl).Perform(WM_NEXTDLGCTL,CtlDir,0);
  end
  else inherited KeyDown(Key,Shift);

  if not (FEditType = etString) then
  begin
    I:=GetInteger;
    case Key of
      VK_PRIOR: if I < FMax then SetInteger(I + FNextPriorStep);
      VK_NEXT: if I > FMin then SetInteger(I - FNextPriorStep);
    end;
  end;
  KeyPressed:=True;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_PAINT: if Focused then Color:=FFocusColor
                else if not Error then Color:=FNoFocusColor else Color:=FErrorColor;
    WM_SETFOCUS: begin
                   SetAlignment(FFocusAlignment);
                   if AutoSelect then SelectAll;
                 end;
    cm_MouseEnter: if FAutoFocus then SetFocus;
    cm_Exit: begin
               SetAlignment(FNoFocusAlignment);
               KeyPressed:=False;
             end;
  end;
  inherited;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.FormatValue;
begin
  Error:=CheckError;
  case FEditType of
    etString: if MaxLength > 0 then Text:=Copy(Text,1,Succ(MaxLength));
    etFloat: SetFloat(GetFloat);
    etInteger,
    etHex,
    etBinary: SetInteger(GetInteger);
  end;
  Perform(WM_PAINT,0,0);
end;
{------------------------------------------------------------------------------}

function TCrossEdit.CheckError: Boolean;
var
  Temp: Double;
begin
  Result:=False;
  if FEditType = etString then Exit;
  Temp:=0;
  Result:=NoValidValue;
  if Result and Assigned(FOnInvalidValue) then FOnInvalidValue(Self)
  else if not Result and FLimitCheck then
  begin
    case FEditType of
      etFloat: Temp:=GetFloat;
      etInteger: Temp:=GetInteger;
      etHex: Temp:=GetInteger;
      etBinary: Temp:=GetInteger;
    end;
    Result:=Temp < FMin;
    if Result and Assigned(FOnLimitMin) then FOnLimitMin(Self)
    else if not Result then
    begin
      Result:=Temp > FMax;
      if Result then if Assigned(FOnLimitMax) then FOnLimitMax(Self);
    end;
  end;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.SetNoFocusAlignment(Value: TAlignment);
begin
  if Value <> FNoFocusAlignment then
  begin
    FNoFocusAlignment:=Value;
    SetAlignment(FNoFocusAlignment);
  end;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.SetAlignment(Value: TAlignment);
begin
  if Value <> Alignment then
  begin
    Alignment:=Value;
    RecreateWnd;
  end;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.SetEditType(Value: TEditType);
begin
  FEditType:=Value;
  case FEditType of
    etBinary: CharSet:=['0','1'];
    etHex: CharSet:=NumChars + ['A'..'F'];
    etInteger: CharSet:=NumChars + ['-'];
    etFloat: CharSet:=NumChars + ['-'] + [DecimalSeparator] + ['E'];
    etString: SetStringCharSet(FStringCharSet);
  end;
  if FEditType <> etString then
  begin
    if NoValidValue then Text:='0';
    FormatValue;
  end;
  SetAlignment(FNoFocusAlignment);
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.SetColor(Index: Integer; Value: TColor);
begin
  case Index of
    0: FFocusColor:=Value;
    1: FNoFocusColor:=Value;
    2: FErrorColor:=Value;
  end;
  Repaint;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.SetStringCharSet(Value: TStringCharSet);
begin
  FStringCharSet:=Value;
  if FEditType = etString then case FStringCharSet of
    scFull: CharSet:=[#0..#255];
    scAlpha: CharSet:=['A'..'Z','a'..'z'];
    scDigit: CharSet:=['0'..'9'];
    scTime: CharSet:=['0'..'9',':'];
    scDate: CharSet:=['0'..'9','.'];
  end;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.SetInteger(Value: Integer);
var
  Len: Integer;
begin
  Len:=MaxLength;
  case FEditType of
    etFloat: Text:=FloatToStrF(Value,ffFixed,FPrecision,FDecimals);
    etInteger: Text:=IntToStr(Value);
    etHex: begin
             if Len < 1 then Len:=4;
             Text:=IntToHex(Value,Len);
             if Length(Text) > Len then
               Text:=Copy(Text,Succ(Length(Text) - Len),Len);
           end;
    etBinary: begin
                if Len < 1 then Len:=8;
                Text:=IntToBin(Value,Len);
              end;
  end;
end;
{------------------------------------------------------------------------------}

procedure TCrossEdit.SetFloat(Value: Double);
begin
  if FEditType = etFloat then Text:=FloatToStrF(Value,ffFixed,FPrecision,FDecimals)
  else SetInteger(Round(Value));
end;
{------------------------------------------------------------------------------}

function TCrossEdit.GetInteger: Integer;
var
  Error: Integer;
begin
  Result:=0;
  if Length(Text) > 0 then case FEditType of
    etFloat,
    etInteger: Val(Text,Result,Error);
    etHex: Result:=StrToInt('$' + Text);
    etBinary: Result:=BinToInt(Text);
  end;
end;
{------------------------------------------------------------------------------}

function TCrossEdit.GetFloat: Double;
begin
  Result:=0.0;
  if (FEditType = etFloat) and not NoValidValue then Result:=StrToFloat(Text);
end;
{------------------------------------------------------------------------------}

function TCrossEdit.NoValidValue: Boolean;
var
  N: Byte;
begin
  Result:=False;
  if FEditType = etString then Exit;
  Text:=Trim(Text);
  if Length(Text) > 0 then
  begin
    for N:=1 to Length(Text) do if not (Text[N] in CharSet) then Result:=True;
    if not Result then if (FEditType = etInteger) or (FEditType = etFloat) then
    begin
      N:=Pos('-',Text);
      if (N > 1) or ((N > 0) and (Length(Text) < 2)) then Result:=True;
    end;
  end
  else Result:=True;
end;
{------------------------------------------------------------------------------}

function TCrossEdit.IntToBin(Value,Digits: Integer): ShortString;
var
  C: Cardinal;
  N: Integer;
begin
  Result:='';
  C:=1;
  if (Digits < 1) or (Digits > 32) then Exit;
  for N:=1 to Digits do
  begin
    if Value and C = C then Result:='1' + Result else Result:='0' + Result;
    C:=C shl 1;
  end;
end;
{------------------------------------------------------------------------------}

function TCrossEdit.BinToInt(const Value: ShortString): Integer;
var
  Add: Integer;
  N,Len: Byte;
begin
  Result:=0;
  Len:=Length(Value);
  if (Len < 1) or (Len > MaxBinary) then Exit;
  Add:=1;
  for N:=Len downto 1 do
  begin
    if not (Value[N] in ['0','1']) then Exit;
    if Value[N] = '1' then Inc(Result,Add);
    Add:=Add shl 1;
  end;
end;
{------------------------------------------------------------------------------}

initialization
end.

